"""
Data Input Output class
"""
# package to add support for multi-language (i18n)
import gettext
# package to handle files/folders and related metadata/operations
import os
# package facilitating Data Frames manipulation
import pandas


class DataInputOutput:
    locale = None

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    @staticmethod
    def fn_add_missing_defaults_to_dict_message(in_dict):
        if 'field delimiter' not in in_dict:
            in_dict['field delimiter'] = os.pathsep
        if 'compression' not in in_dict:
            in_dict['compression'] = 'infer'
        return in_dict

    def fn_build_feedback_for_logger(self, operation_details):
        messages = {}
        if operation_details['operation'] == 'load':
            files_counted = str(operation_details['files counted'])
            messages = {
                'failed': self.locale.gettext(
                    'Error encountered on loading Pandas Data Frame '
                    + 'from {file_type} file type (see below)')
                    .replace('{file_type}',  operation_details['format'].upper()),
                'success': self.locale.gettext(
                    'All {files_counted} files of type {file_type} '
                    + 'successfully added to a Pandas Data Frame')
                    .replace('{files_counted}', files_counted)
                    .replace('{file_type}', operation_details['format'].upper())
            }
        elif operation_details['operation'] == 'save':
            messages = {
                'failed': self.locale.gettext(
                    'Error encountered on saving Pandas Data Frame '
                    + 'into a {file_type} file type (see below)')
                    .replace('{file_type}',  operation_details['format'].upper()),
                'success': self.locale.gettext(
                    'Pandas Data Frame has just been saved to file "{file_name}", '
                    + 'considering {file_type} as file type')
                    .replace('{file_name}',operation_details['name'])
                    .replace('{file_type}',  operation_details['format'].upper()),
            }
        return messages

    def fn_file_operation_logger(self, local_logger, in_logger_dict):
        messages = self.fn_build_feedback_for_logger(in_logger_dict)
        if in_logger_dict['error details'] is None:
            local_logger.info(messages['success'])
        else:
            local_logger.error(messages['failed'])
            local_logger.error(in_logger_dict['error details'])

    def fn_load_file_into_data_frame(self, in_logger, timer, in_dict):
        timer.start()
        if self.fn_store_data_frame_to_file_validation(in_logger, in_dict):
            in_dict = self.fn_add_missing_defaults_to_dict_message(in_dict)
            in_dict.update({'operation': 'load'})
            in_dict = self.fn_pack_dict_message(in_dict, in_dict['file list'])
            in_dict = self.fn_internal_load_csv_file_into_data_frame(in_dict)
            in_dict = self.fn_internal_load_excel_file_into_data_frame(in_dict)
            in_dict = self.fn_internal_load_json_file_into_data_frame(in_dict)
            in_dict = self.fn_internal_load_pickle_file_into_data_frame(in_dict)
            self.fn_file_operation_logger(in_logger, in_dict)
        timer.stop()
        return in_dict['out data frame']

    @staticmethod
    def fn_internal_load_csv_file_into_data_frame(in_dict):
        if in_dict['format'].lower() == 'csv':
            try:
                in_dict['out data frame'] = pandas.concat(
                    [pandas.read_csv(filepath_or_buffer=crt_file,
                                     delimiter=in_dict['field delimiter'],
                                     cache_dates=True,
                                     index_col=None,
                                     memory_map=True,
                                     low_memory=False,
                                     encoding='utf-8',
                                     ) for crt_file in in_dict['files list']])
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_load_excel_file_into_data_frame(in_dict):
        if in_dict['format'].lower() == 'excel':
            try:
                in_dict['out data frame'] = pandas.concat(
                    [pandas.read_excel(io=crt_file,
                                       verbose=True,
                                       ) for crt_file in in_dict['files list']])
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_load_json_file_into_data_frame(in_dict):
        if in_dict['format'].lower() == 'json':
            try:
                in_dict['out data frame'] = pandas.concat(
                    [pandas.read_json(filepath_or_buffer=crt_file,
                                      compression=in_dict['compression'],
                                      ) for crt_file in in_dict['files list']])
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_load_pickle_file_into_data_frame(in_dict):
        if in_dict['format'].lower() == 'pickle':
            try:
                in_dict['out data frame'] = pandas.concat(
                    [pandas.read_pickle(path=crt_file,
                                        compression=in_dict['compression'],
                                        ) for crt_file in in_dict['files list']])
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_store_data_frame_to_csv_file(in_dict):
        if in_dict['format'].lower() == 'csv':
            try:
                in_dict['in data frame'].to_csv(path_or_buf=in_dict['name'],
                                                sep=in_dict['field delimiter'],
                                                header=True,
                                                index=False,
                                                encoding='utf-8')
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_store_data_frame_to_excel_file(in_dict):
        if in_dict['format'].lower() == 'excel':
            try:
                in_dict['in data frame'].to_excel(excel_writer=in_dict['name'],
                                                  engine='xlsxwriter',
                                                  freeze_panes=(1, 1),
                                                  encoding='utf-8',
                                                  index=False,
                                                  verbose=True)
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_store_data_frame_to_json_file(in_dict):
        if in_dict['format'].lower() == 'json':
            try:
                in_dict['in data frame'].to_json(path_or_buf=in_dict['name'],
                                                 compression=in_dict['compression'])
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_store_data_frame_to_pickle_file(in_dict):
        if in_dict['format'].lower() == 'pickle':
            try:
                in_dict['in data frame'].to_pickle(path=in_dict['name'],
                                                   compression=in_dict['compression'])
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_pack_dict_message(in_dict, in_file_list):
        return {
            'compression': in_dict['compression'],
            'field delimiter': in_dict['field delimiter'],
            'files list': in_file_list,
            'files counted': len(in_file_list),
            'error details': None,
            'format': in_dict['format'],
            'name': in_dict['name'],
            'in data frame': None,
            'operation': in_dict['operation'],
            'out data frame': None,
        }

    def fn_store_data_frame_to_file(self, in_logger, timer, in_data_frame, in_dict):
        timer.start()
        if self.fn_store_data_frame_to_file_validation(in_logger, in_dict):
            in_dict = self.fn_add_missing_defaults_to_dict_message(in_dict)
            in_dict.update({'operation': 'save'})
            in_dict = self.fn_pack_dict_message(in_dict, [])
            in_dict.update({'in data frame': in_data_frame})
            in_dict = self.fn_internal_store_data_frame_to_csv_file(in_dict)
            in_dict = self.fn_internal_store_data_frame_to_excel_file(in_dict)
            in_dict = self.fn_internal_store_data_frame_to_json_file(in_dict)
            in_dict = self.fn_internal_store_data_frame_to_pickle_file(in_dict)
            self.fn_file_operation_logger(in_logger, in_dict)
        timer.stop()

    def fn_store_data_frame_to_file_validation(self, local_logger, in_file_details):
        given_format_is_implemented = False
        if 'format' in in_file_details:
            implemented_file_formats = ['csv', 'excel', 'json', 'pickle']
            given_format = in_file_details['format'].lower()
            given_format_is_implemented = True
            if given_format not in implemented_file_formats:
                given_format_is_implemented = False
                local_logger.error(self.locale.gettext(
                    'File "format" attribute has a value of "{format_value}" '
                    + 'which is not among currently implemented values: '
                    + '"{implemented_file_formats}", '
                    + 'therefore desired file operation is not possible')
                                   .replace('{format_value}', given_format)
                                   .replace('{implemented_file_formats}',
                                            '", "'.join(implemented_file_formats)))
        else:
            local_logger.error(self.locale.gettext(
                    'File "format" attribute is mandatory in the file setting, but missing, '
                    + 'therefore desired file operation is not possible'))
        return given_format_is_implemented



"""
facilitates File Operations
"""
# package to handle date and times
from datetime import datetime
# package to add support for multi-language (i18n)
import gettext
# package to get ability to search file recursively
import glob
# package to use for checksum calculations (in this file)
import hashlib
# package to handle json files
import json
# package to facilitate operating system operations
import os
# package to facilitate os path manipulations
import pathlib
# package regular expressions
import re


class FileOperations:
    timestamp_format = '%Y-%m-%d %H:%M:%S.%f %Z'
    locale = None

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    def fn_build_file_list(self, local_logger, timer, given_input_file):
        timer.start()
        if re.search(r'(\*|\?)*', given_input_file):
            local_logger.debug(self.locale.gettext('File matching pattern identified'))
            parent_directory = os.path.dirname(given_input_file)
            # loading from a specific folder all files matching a given pattern into a file list
            relevant_files_list = self.fn_build_relevant_file_list(
                local_logger, parent_directory, given_input_file)
        else:
            local_logger.debug(self.locale.gettext('Specific file name provided'))
            relevant_files_list = [given_input_file]
        timer.stop()
        return relevant_files_list

    def fn_build_relevant_file_list(self, local_logger, in_folder, matching_pattern):
        folder_parts = pathlib.Path(matching_pattern).parts
        search_pattern = folder_parts[(len(folder_parts)-1)]
        local_logger.info(
            self.locale.gettext('Listing all files within {in_folder} folder '
                                + 'looking for {matching_pattern} as matching pattern')
                    .replace('{in_folder}', in_folder)
                    .replace('{matching_pattern}', search_pattern))
        list_files = []
        if os.path.isdir(in_folder):
            list_files = glob.glob(matching_pattern)
            file_counter = len(list_files)
            local_logger.info(self.locale.ngettext(
                '{files_counted} file from {in_folder} folder identified',
                '{files_counted} files from {in_folder} folder identified', file_counter)
                              .replace('{files_counted}', str(file_counter))
                              .replace('{in_folder}', in_folder))
        else:
            local_logger.error(self.locale.gettext('Folder {folder_name} does not exist')
                               .replace('{folder_name}', in_folder))
        return list_files

    def fn_get_file_content(self, in_file_handler, in_file_type):
        if in_file_type == 'json':
            try:
                json_interpreted_details = json.load(in_file_handler)
                print(datetime.utcnow().strftime(self.timestamp_format) + '- ' +
                      self.locale.gettext('JSON structure interpreted'))
                return json_interpreted_details
            except Exception as e:
                print(datetime.utcnow().strftime(self.timestamp_format) + '- ' +
                      self.locale.gettext('Error encountered when trying to interpret JSON'))
                print(e)
        elif in_file_type == 'raw':
            raw_interpreted_file = in_file_handler.read()
            print(datetime.utcnow().strftime(self.timestamp_format) + '- ' +
                  self.locale.gettext('Entire file content read'))
            return raw_interpreted_file
        else:
            print(datetime.utcnow().strftime(self.timestamp_format) + '- ' +
                  self.locale.gettext('Unknown file type provided, '
                                      + 'expected either "json" or "raw" but got {in_file_type}')
                  .replace('{in_file_type}', in_file_type))

    @staticmethod
    def fn_get_file_dates(file_to_evaluate):
        return {
            'created': datetime.fromtimestamp(os.path.getctime(file_to_evaluate)),
            'last modified': datetime.fromtimestamp(os.path.getmtime(file_to_evaluate)),
        }

    def fn_get_file_simple_statistics(self, file_to_evaluate):
        file_date_time = self.fn_get_file_dates(file_to_evaluate)
        return {
            'date when created': datetime.strftime(file_date_time['created'],
                                                   self.timestamp_format).strip(),
            'date when last modified': datetime.strftime(file_date_time['last modified'],
                                                         self.timestamp_format).strip(),
            'size [bytes]': os.path.getsize(file_to_evaluate),
        }

    def fn_get_file_statistics(self, file_to_evaluate):
        file_statistics = self.fn_get_file_simple_statistics(file_to_evaluate)
        try:
            file_handler = open(file=file_to_evaluate, mode='r', encoding='mbcs')
        except UnicodeDecodeError:
            file_handler = open(file=file_to_evaluate, mode='r', encoding='utf-8')
        file_content = file_handler.read().encode()
        file_handler.close()
        file_statistics['MD5 Checksum'] = hashlib.md5(file_content).hexdigest()
        file_statistics['SHA1 Checksum'] = hashlib.sha1(file_content).hexdigest()
        file_statistics['SHA224 Checksum'] = hashlib.sha224(file_content).hexdigest()
        file_statistics['SHA256 Checksum'] = hashlib.sha256(file_content).hexdigest()
        file_statistics['SHA384 Checksum'] = hashlib.sha384(file_content).hexdigest()
        file_statistics['SHA512 Checksum'] = hashlib.sha512(file_content).hexdigest()
        return file_statistics

    def fn_get_file_datetime_verdict(self, local_logger, file_to_evaluate,
                                     created_or_modified, reference_datetime):
        implemented_choices = ['created', 'last modified']
        verdict = self.locale.gettext('unknown')
        file_date_time = self.fn_get_file_dates(file_to_evaluate)
        if created_or_modified in implemented_choices:
            which_datetime = file_date_time.get(created_or_modified)
            verdict = self.locale.gettext('older')
            if which_datetime > reference_datetime:
                verdict = self.locale.gettext('newer')
            elif which_datetime == reference_datetime:
                verdict = self.locale.gettext('same')
            str_file_datetime = datetime.strftime(which_datetime, self.timestamp_format).strip()
            str_reference = datetime.strftime(reference_datetime, self.timestamp_format).strip()
            local_logger.debug(self.locale.gettext(
                    'File "{file_name}" which has the {created_or_modified} datetime '
                    + 'as "{file_datetime}" vs. "{reference_datetime}" '
                    + 'has a verdict = {verdict}')
                              .replace('{file_name}', str(file_to_evaluate))
                              .replace('{created_or_modified}',
                                       self.locale.gettext(created_or_modified))
                              .replace('{reference_datetime}', str_reference)
                              .replace('{file_datetime}', str_file_datetime)
                              .replace('{verdict}', verdict))
        else:
            local_logger.error(self.locale.gettext(
                    'Unknown file datetime choice, '
                    + 'expected is one of the following choices "{implemented_choices}" '
                    + 'but provided was "{created_or_modified}"...')
                               .replace('{implemented_choices}', '", "'.join(implemented_choices))
                               .replace('{created_or_modified}', created_or_modified))
        return verdict

    def fn_move_files(self, local_logger, timer, file_names, destination_folder):
        timer.start()
        resulted_files = []
        for current_file in file_names:
            source_folder = os.path.dirname(current_file)
            new_file_name = current_file.replace(source_folder, destination_folder)
            if os.path.isfile(new_file_name):
                local_logger.warning(self.locale.gettext('File {file_name} will be overwritten')
                                     .replace('{file_name}', current_file))
                os.replace(current_file, new_file_name)
                local_logger.info(self.locale.gettext(
                    'File {file_name} has just been been overwritten as {new_file_name}')
                                     .replace('{file_name}', current_file)
                                     .replace('{new_file_name}', current_file))
            else:
                local_logger.info(self.locale.gettext(
                    'File {file_name} will be renamed as {new_file_name}')
                                     .replace('{file_name}', current_file)
                                     .replace('{new_file_name}', current_file))
                os.rename(current_file, new_file_name)
                local_logger.info(self.locale.gettext(
                    'File {file_name} has just been renamed as {new_file_name}')
                                     .replace('{file_name}', current_file)
                                     .replace('{new_file_name}', current_file))
            resulted_files.append(str(new_file_name))
        timer.stop()
        return resulted_files

    def fn_open_file_and_get_content(self, input_file, content_type='json'):
        if os.path.isfile(input_file):
            with open(input_file, 'r') as file_handler:
                print(datetime.utcnow().strftime(self.timestamp_format) + '- ' +
                      self.locale.gettext('File {file_name} has just been opened')
                      .replace('{file_name}', str(input_file)))
                return self.fn_get_file_content(file_handler, content_type)
        else:
            print(datetime.utcnow().strftime(self.timestamp_format) + '- ' +
                  self.locale.gettext('File {file_name} does not exist')
                  .replace('{file_name}', str(input_file)))

    def fn_store_file_statistics(self, local_logger, timer, file_name, file_meaning):
        timer.start()
        list_file_names = [file_name]
        if type(file_name) == list:
            list_file_names = file_name
        for current_file_name in list_file_names:
            dt_when_last_modified = 'date when last modified'
            file_statistics = str(self.fn_get_file_statistics(current_file_name))\
                .replace('date when created', self.locale.gettext('date when created')) \
                .replace(dt_when_last_modified, self.locale.gettext(dt_when_last_modified)) \
                .replace('size [bytes]', self.locale.gettext('size [bytes]')) \
                .replace('Checksum', self.locale.gettext('Checksum'))
            local_logger.info(self.locale.gettext(
                'File "{file_name}" has the following characteristics: {file_statistics}')
                              .replace('{file_meaning}', file_meaning)
                              .replace('{file_name}', current_file_name)
                              .replace('{file_statistics}', file_statistics))
        timer.stop()



"""
CommandLineArgumentManagement - library to manage input parameters from command line

This library allows handling pre-configured arguments to be received from command line and use them
to call the main package functions
"""
# package to handle arguments from command line
import argparse
# package to add support for multi-language (i18n)
import gettext
# package to facilitate operating system operations
import os


class CommandLineArgumentsManagement:
    locale = None

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    def listing_parameter_values(self, in_logger, timer, title, in_config, given_parameter_values):
        timer.start()
        in_logger.info('=' * 50)
        in_logger.info(self.locale.gettext('{application_name} has started')
                       .replace('{application_name}', title))
        in_logger.info('~' * 50)
        in_logger.info(self.locale.gettext('Overview of input parameter given values'))
        in_logger.info('~' * 50)
        parameter_values_dictionary = given_parameter_values.__dict__
        for input_key, attributes in in_config.items():
            # checking first if short key was provided, otherwise consider longer
            if input_key in parameter_values_dictionary:
                key_value_to_consider = input_key
            else:
                key_value_to_consider = attributes['option_long'].replace('-', '_')
            # having the key consider we determine the value of the current parameter
            value_to_consider = parameter_values_dictionary[key_value_to_consider]
            # we build the parameter feedback considering "option_description"
            # and replacing %s with parameter value
            feedback = self.locale.gettext(attributes['option_description']) \
                .replace('%s', str(value_to_consider))
            # we finally write the feedback to logger
            in_logger.info(feedback)
        in_logger.info('~' * 50)
        timer.stop()

    def parse_arguments(self, configuration_details):
        parser = argparse.ArgumentParser()
        for input_key, attributes in configuration_details.items():
            action_value = self.translate_default_to_action(attributes['default_value'])
            if action_value is None:
                parser.add_argument('-' + input_key, '--' + attributes['option_long'],
                                    required=attributes['option_required'],
                                    default=attributes['default_value'],
                                    help=attributes['option_sample_value'])
            else:
                parser.add_argument('-' + input_key, '--' + attributes['option_long'],
                                    required=attributes['option_required'],
                                    default=attributes['default_value'],
                                    action=action_value)
        parser.add_argument('-v', '--verbose', required=False, default=False, action='store_true')
        return parser.parse_args()

    @staticmethod
    def translate_default_to_action(given_default_value):
        if given_default_value is True:
            return 'store_true'
        elif given_default_value is False:
            return 'store_false'
        else:
            return None



"""
Data Manipulation class
"""
# package to add support for multi-language (i18n)
import gettext
# package to handle files/folders and related metadata/operations
import os


class DataManipulator:
    locale = None

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    def fn_add_and_shift_column(self, local_logger, timer, input_data_frame, input_details: list):
        evr = 'Empty Values Replacement'
        for crt_dict in input_details:
            timer.start()
            input_data_frame[crt_dict['New Column']] = input_data_frame[crt_dict['Original Column']]
            col_offset = self.fn_set_shifting_value(crt_dict)
            input_data_frame[crt_dict['New Column']] = \
                input_data_frame[crt_dict['New Column']].shift(col_offset)
            input_data_frame[crt_dict['New Column']] = \
                input_data_frame[crt_dict['New Column']].apply(lambda x: str(x)
                                                               .replace('nan', str(crt_dict[evr]))
                                                               .replace('.0', ''))
            local_logger.info(self.locale.gettext(
                'A new column named "{new_column_name}" as copy from "{original_column}" '
                + 'then shifted by {shifting_rows} to relevant data frame '
                + '(filling any empty value as {empty_values_replacement})')
                              .replace('{new_column_name}', crt_dict['New Column'])
                              .replace('{original_column}', crt_dict['Original Column'])
                              .replace('{shifting_rows}', str(col_offset))
                              .replace('{empty_values_replacement}',
                                       str(crt_dict['Empty Values Replacement'])))
            timer.stop()
        return input_data_frame

    @staticmethod
    def fn_add_minimum_and_maximum_columns_to_data_frame(input_data_frame, dict_expression):
        grouped_df = input_data_frame.groupby(dict_expression['group_by']) \
            .agg({dict_expression['calculation']: ['min', 'max']})
        grouped_df.columns = ['_'.join(col).strip() for col in grouped_df.columns.values]
        grouped_df = grouped_df.reset_index()
        if 'map' in dict_expression:
            grouped_df.rename(columns=dict_expression['map'], inplace=True)
        return grouped_df

    def fn_apply_query_to_data_frame(self, local_logger, timer, input_data_frame, extract_params):
        timer.start()
        query_expression = ''
        generic_pre_feedback = self.locale.gettext('Will retain only values {filter_type} '
                                                   + '"{filter_values}" within the field '
                                                   + '"{column_to_filter}"') \
            .replace('{column_to_filter}', extract_params['column_to_filter'])
        if extract_params['filter_to_apply'] == 'equal':
            local_logger.debug(generic_pre_feedback
                               .replace('{filter_type}', self.locale.gettext('equal with'))
                               .replace('{filter_values}', extract_params['filter_values']))
            query_expression = '`' + extract_params['column_to_filter'] + '` == "' \
                               + extract_params['filter_values'] + '"'
        elif extract_params['filter_to_apply'] == 'different':
            local_logger.debug(generic_pre_feedback
                               .replace('{filter_type}', self.locale.gettext('different than'))
                               .replace('{filter_values}', extract_params['filter_values']))
            query_expression = '`' + extract_params['column_to_filter'] + '` != "' \
                               + extract_params['filter_values'] + '"'
        elif extract_params['filter_to_apply'] == 'multiple_match':
            multiple_values = '["' + '", "'.join(extract_params['filter_values'].values()) + '"]'
            local_logger.debug(generic_pre_feedback
                               .replace('{filter_type}',
                                        self.locale.gettext('matching any of these values'))
                               .replace('{filter_values}', multiple_values))
            query_expression = '`' + extract_params['column_to_filter'] + '` in ' + multiple_values
        local_logger.debug(self.locale.gettext('Query expression to apply is: {query_expression}')
                           .replace('{query_expression}', query_expression))
        input_data_frame.query(query_expression, inplace=True)
        timer.stop()
        return input_data_frame

    def fn_filter_data_frame_by_index(self, local_logger, in_data_frame, filter_rule):
        reference_expression = filter_rule['Query Expression for Reference Index']
        index_current = in_data_frame.query(reference_expression, inplace=False)
        local_logger.info(self.locale.gettext(
            'Current index has been determined to be {index_current_value}')
                          .replace('{index_current_value}', str(index_current.index)))
        if str(index_current.index) != "Int64Index([], dtype='int64')" \
                and 'Deviation' in filter_rule:
            in_data_frame = self.fn_filter_data_frame_by_index_internal(local_logger, {
                'data frame': in_data_frame,
                'deviation': filter_rule['Deviation'],
                'index': index_current.index,
            })
        return in_data_frame

    def fn_filter_data_frame_by_index_internal(self, local_logger, in_dict):
        in_data_frame = in_dict['data_frame']
        for deviation_type in in_dict['deviation']:
            deviation_number = in_dict['deviation'][deviation_type]
            index_to_apply = in_dict['index']
            if deviation_type == 'Lower':
                index_to_apply -= deviation_number
                in_data_frame = in_data_frame[in_dict['index'] >= index_to_apply[0]]
            elif deviation_type == 'Upper':
                index_to_apply += deviation_number
                in_data_frame = in_data_frame[in_dict['index'] <= index_to_apply[0]]
            local_logger.info(self.locale.gettext(
                '{deviation_type} Deviation Number is {deviation_number} '
                + 'to be applied to Current index, became {index_to_apply}')
                              .replace('{deviation_type}', deviation_type)
                              .replace('{deviation_number}', str(deviation_number))
                              .replace('{index_to_apply}', str(index_to_apply)))
        return in_dict['data_frame']

    @staticmethod
    def fn_get_column_index_from_data_frame(data_frame_columns, column_name_to_identify):
        column_index_to_return = 0
        for ndx, column_name in enumerate(data_frame_columns):
            if column_name == column_name_to_identify:
                column_index_to_return = ndx
        return column_index_to_return

    @staticmethod
    def fn_get_first_and_last_column_value_from_data_frame(in_data_frame, in_column_name):
        return {
            'first': in_data_frame.iloc[0][in_column_name],
            'last': in_data_frame.iloc[(len(in_data_frame) - 1)][in_column_name],
        }

    @staticmethod
    def fn_set_shifting_value(in_dict):
        offset_sign = 1
        if in_dict['Direction'] == 'up':
            offset_sign = -1
        return offset_sign * in_dict['Deviation']



"""
localization_compile - facilitates localization file compilation (from .po to .mo)
"""
import os
import pathlib

# specific to this project
from localizations_common import LocalizationsCommon


class CustomizedLocalizationCompiling(LocalizationsCommon):
    localisation_compilation_is_required = False

    def evaluate_compilation_necessity(self, in_list_localisation_source_files):
        list_size = len(in_list_localisation_source_files)
        self.operation_is_required = False
        file_list_paring_complete = False
        file_counter = 0
        domains_to_compile = []
        compiling_files_counter = 0
        while not file_list_paring_complete:
            source_localisation_file = in_list_localisation_source_files[file_counter]
            folder_parts = pathlib.PurePath(source_localisation_file).parts
            current_locale = folder_parts[(len(folder_parts) - 3)]
            fn_dict = {
                'destination': source_localisation_file.replace('.po', '.mo'),
                'counter': file_counter,
                'locale': current_locale,
                'source': source_localisation_file,
                'source file type name': 'source',
                'destination operation name': 'compilation',
            }
            operation_check_result, operation_to_execute = self.check_file_pairs(fn_dict)
            if operation_check_result:
                domains_to_compile.append(compiling_files_counter)
                domains_to_compile[compiling_files_counter] = {
                    'input-file': fn_dict['source'],
                    'operation': 'compile_catalog',
                    'operation final flags': ' --statistics',
                    'output-file': fn_dict['destination'],
                    'locale': fn_dict['locale'],
                }
                compiling_files_counter += 1
            file_counter += 1
            file_list_paring_complete = self.file_counter_limit(file_counter, list_size)
        return domains_to_compile


my_class = CustomizedLocalizationCompiling()

locale_source_files = my_class.get_project_localisation_source_files('po')
operation_locale_dict = my_class.evaluate_compilation_necessity(locale_source_files)
my_class.operate_localisation_files(operation_locale_dict)



"""
LoggingNeeds

This library sets all required parameters for a rotating LOG file
"""
from datetime import datetime
# package for Log management
import logging
import logging.handlers as handlers


class LoggingNeeds:
    logger = None

    def initiate_logger(self, logger_file_name, logger_internal_name):
        # initiate Logging
        self.logger = logging.getLogger(logger_internal_name)
        # set logging level to desired level only if specified
        if logger_file_name == 'None':
            self.logger.setLevel(logging.NOTSET)
        else:
            self.logger.setLevel(logging.DEBUG)
            # defining log file and setting rotating logic
            log_handler = handlers.TimedRotatingFileHandler(logger_file_name,
                                                            when='h',
                                                            interval=1,
                                                            backupCount=5,
                                                            encoding='utf-8',
                                                            utc=True)
            # ensure timestamps are reported as UTC time-zone
            my_converter = lambda x, y: datetime.utcnow().timetuple()
            logging.Formatter.converter = my_converter
            # Here we define our formatter
            log_formatter = logging.Formatter(
                    '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            # Here we set our logHandler's formatter
            log_handler.setFormatter(log_formatter)
            # pairing the handler with logging
            self.logger.addHandler(log_handler)



"""
main - entry point of the package

This file is connecting to a Tableau Server and publishes a local HYPER file
measuring time elapsed (performance)
"""

# package to facilitate operating system project_locale detection
import locale
# package to handle files/folders and related metadata/operations
import os
# package to facilitate multiple operation system operations
import platform

# Custom classes specific to this package
from project_locale.localizations_common import LocalizationsCommon
from tableau_hyper_management.ProjectNeeds import ProjectNeeds
from tableau_hyper_management.TableauServerCommunicator import TableauServerCommunicator
# get current script name
#SCRIPT_NAME = os.path.basename(__file__).replace('.py', '')
SCRIPT_NAME = 'publisher'

# main execution logic
if __name__ == '__main__':
    # instantiate Localizations Common class
    class_lc = LocalizationsCommon()
    # ensure all compiled localization files are in place (as needed for localized messages later)
    class_lc.run_localization_compile()
    # establish localization language to use
    language_to_use = class_lc.get_region_language_to_use_from_operating_system()
    # instantiate Extractor Specific Needs class
    class_pn = ProjectNeeds(SCRIPT_NAME, language_to_use)
    # load application configuration (inputs are defined into a json file)
    class_pn.load_configuration()
    # initiate Logging sequence
    class_pn.initiate_logger_and_timer()
    # reflect title and input parameters given values in the log
    class_pn.class_clam.listing_parameter_values(
        class_pn.class_ln.logger, class_pn.timer, 'Tableau Data Source Publisher',
        class_pn.config['input_options'][SCRIPT_NAME], class_pn.parameters)
    relevant_files_list = class_pn.class_fo.fn_build_file_list(
            class_pn.class_ln.logger, class_pn.timer, class_pn.parameters.input_file)
    # log file statistic details
    class_pn.class_fo.fn_store_file_statistics(
            class_pn.class_ln.logger, class_pn.timer, relevant_files_list, 'Input')
    # get the secrets from provided file
    credentials = class_pn.class_fo.fn_open_file_and_get_content(
            class_pn.parameters.input_credentials_file, 'json')
    credentials_dict = credentials['Credentials']['LDAP']['Production']['Default']
    # instantiate main library that ensure Tableau Server communication
    c_tsc = TableauServerCommunicator()
    # initiate Tableau Server connection
    c_tsc.connect_to_tableau_server(class_pn.class_ln.logger, class_pn.timer, {
        'Tableau Server': class_pn.parameters.tableau_server,
        'Tableau Site': class_pn.parameters.tableau_site,
        'Username': credentials_dict['Username'],
        'Password': credentials_dict['Password'],
    })
    # identify relevant project(s) based on given name
    list_project_details = c_tsc.load_tableau_project_ids(
            class_pn.class_ln.logger, class_pn.timer,
            [class_pn.parameters.tableau_project], 'JustOnesMentioned')
    # check if a single project has been identified and if so proceed with publishing
    if c_tsc.is_publishing_possible(
            class_pn.class_ln.logger, class_pn.parameters.tableau_project, list_project_details):
        # perform the publishing of data source
        c_tsc.publish_data_source_to_tableau_server(class_pn.class_ln.logger, class_pn.timer, {
            'Project ID': list_project_details[0],
            'Tableau Extract File': class_pn.parameters.input_file,
            'Publishing Mode': class_pn.parameters.publishing_mode,
        })
    # disconnect from Tableau Server
    c_tsc.disconnect_from_tableau_server(class_pn.class_ln.logger, class_pn.timer)
    # just final message
    class_pn.class_bn.fn_final_message(
            class_pn.class_ln.logger, class_pn.parameters.output_log_file,
            class_pn.timer.timers.total(SCRIPT_NAME))



"""
TableauServerCommunication - a Tableau Server Client wrapper

This library facilitates publishing data source to Tableau Server
"""
# package to add support for multi-language (i18n)
import gettext
# package to handle files/folders and related metadata/operations
import os
# package to ensure communication with Tableau Server
import tableauserverclient as tsc
# Path manager
from pathlib import Path


class TableauServerCommunicator:
    req_opts = tsc.RequestOptions(pagesize=1000)
    tableau_server = None
    locale = None

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    def connect_to_tableau_server(self, local_logger, timer, in_connection):
        timer.start()
        local_logger.debug(self.locale.gettext(
            'About to connect to the Tableau Server with URL: {tableau_server}, '
            + 'Site {tableau_site} with Username = {user_name}')
                           .replace('{tableau_server}', in_connection['Tableau Server'])
                           .replace('{tableau_site}', in_connection['Tableau Site'])
                           .replace('{user_name}', in_connection['Username']))
        self.tableau_server = tsc.Server(in_connection['Tableau Server'], True)
        tableau_auth = tsc.TableauAuth(
            in_connection['Username'], in_connection['Password'], in_connection['Tableau Site'])
        self.tableau_server.auth.sign_in(tableau_auth)
        local_logger.debug(self.locale.gettext(
            'Connection to the Tableau Server has been established successfully!'))
        timer.stop()

    def disconnect_from_tableau_server(self, local_logger, timer):
        timer.start()
        self.tableau_server.auth.sign_out()
        local_logger.debug(self.locale.gettext(
            'Connection to the Tableau Server has been terminated!'))
        timer.stop()

    @staticmethod
    def get_project_relevancy(relevant_projects_to_filter, filtering_type, current_project):
        relevant = 0
        if filtering_type == 'JustOnesMentioned':
            if current_project.name.replace(chr(8211), chr(45)) in relevant_projects_to_filter:
                relevant = 1
        elif filtering_type == 'OnesMentionedMarked':
            relevant = 2
        elif filtering_type == 'All':
            relevant = 99
        return relevant

    def is_publishing_possible(self, local_logger, relevant_project_name, relevant_project_ids):
        publish_possible = False
        int_found_projects = len(relevant_project_ids)
        if int_found_projects == 0:
            local_logger.error(self.locale.gettext(
                'No project with provided name "{project_name}" has been found')
                               .replace('{project_name}', relevant_project_name))
            self.no_publishing_feedback()
        elif int_found_projects > 1:
            local_logger.error(f'There are {str(int_found_projects)} projects with provided name "'
                               + relevant_project_name + ' but a unique identifier is expected')
            self.no_publishing_feedback()
        else:
            publish_possible = True
            local_logger.error(self.locale.gettext(
                'A single project identifier was found "{project_guid}", '
                + 'so I will proceed with publishing provided file there')
                               .replace('{project_guid}', relevant_project_ids[0]))
            local_logger.info(self.locale.gettext('Stay tuned for the confirmation'))
        return publish_possible

    def load_tableau_project_ids(self, local_logger, timer, in_projects_to_filter, in_filter_type):
        timer.start()
        project_items, pagination_item = self.tableau_server.projects.get(req_options=self.req_opts)
        local_logger.info(self.locale.gettext(
            'Reading list of all projects available has been completed'))
        local_logger.info(self.locale.gettext(
            'A number of {projects_counted} projects have been identified')
                          .replace('{projects_counted}', str(len(project_items))))
        timer.stop()
        timer.start()
        dictionary_project_ids = []
        project_counter = 0
        for project_current in project_items:
            relevant = self.get_project_relevancy(
                in_projects_to_filter, in_filter_type, project_current)
            if relevant >= 1:
                dictionary_project_ids.append(project_counter)
                dictionary_project_ids[project_counter] = project_current.id
                project_counter = project_counter + 1
        local_logger.info(self.locale.gettext(
            'Retaining the projects according to filtering type provided ({filter_type}) '
            + 'has been completed')
                          .replace('{filter_type}', in_filter_type))
        local_logger.info(self.locale.gettext(
            'A number of {projects_counted} projects have been identified')
                          .replace('{projects_counted}', str(len(dictionary_project_ids))))
        timer.stop()
        return dictionary_project_ids

    def no_publishing_feedback(self, local_logger):
        local_logger.debug(self.locale.gettext('No publishing action will take place!'))
        local_logger.debug(self.locale.gettext(
            'Check your input parameter values for accuracy and try again!'))

    def publish_data_source_to_tableau_server(self, local_logger, timer, publish_details):
        timer.start()
        local_logger.info(self.locale.gettext('About to start publishing'))
        data_source_name = Path(publish_details['Tableau Extract File'])\
            .name.replace('.hyper', '') + " Extract"
        project_data_source = tsc.DatasourceItem(publish_details['Project ID'], data_source_name)
        self.tableau_server.datasources.publish(
            project_data_source, publish_details['Tableau Extract File'],
            publish_details['Publishing Mode'])
        local_logger.info(self.locale.gettext('Publishing completed successfully!'))
        timer.stop()



"""
Class Converter Specific Needs

Handling specific needs for Extractor script
"""
# useful methods to measure time performance by small pieces of code
from codetiming import Timer
# package to add support for multi-language (i18n)
import gettext
# package to facilitate operating system operations
import os
# package to facilitate common operations
from .BasicNeeds import BasicNeeds
from .CommandLineArgumentsManagement import CommandLineArgumentsManagement
from .DataInputOutput import DataInputOutput
from .FileOperations import FileOperations
from .LoggingNeeds import LoggingNeeds


class ProjectNeeds:
    class_bn = None
    class_clam = None
    class_dio = None
    class_fo = None
    config = None
    locale = None
    parameters = None
    script = None
    timer = None

    def __init__(self, destination_script, default_language='en_US'):
        self.script = destination_script
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[default_language], fallback=True)
        # instantiate Basic Needs class
        self.class_bn = BasicNeeds(default_language)
        # instantiate File Operations class
        self.class_fo = FileOperations(default_language)
        # instantiate Command Line Arguments class
        self.class_clam = CommandLineArgumentsManagement(default_language)
        # instantiate Data Manipulator class, useful to manipulate data frames
        self.class_dio = DataInputOutput(default_language)
        # instantiate Logger class
        self.class_ln = LoggingNeeds()

    def fn_check_inputs_specific(self, input_parameters):
        if self.script == 'converter':
            self.class_bn.fn_validate_single_value(
                    os.path.dirname(input_parameters.output_file), 'folder')
        if self.script == 'publisher':
            self.class_bn.fn_validate_single_value(
                    input_parameters.input_file, 'file')
            self.class_bn.fn_validate_single_value(
                    input_parameters.input_credentials_file, 'file')
            self.class_bn.fn_validate_single_value(
                    input_parameters.tableau_server, 'url')

    def initiate_logger_and_timer(self):
        # initiate logger
        self.class_ln.initiate_logger(self.parameters.output_log_file, self.script)
        # initiate localization specific for this script
        # define global timer to use
        self.timer = Timer(self.script,
                           text=self.locale.gettext('Time spent is {seconds}'),
                           logger=self.class_ln.logger.debug)

    def load_configuration(self):
        # load application configuration (inputs are defined into a json file)
        ref_folder = os.path.dirname(__file__).replace('tableau_hyper_management', 'config')
        config_file = os.path.join(ref_folder, 'tableau-hyper-management.json').replace('\\', '/')
        self.config = self.class_fo.fn_open_file_and_get_content(config_file)
        # get command line parameter values
        self.parameters = self.class_clam.parse_arguments(self.config['input_options'][self.script])
        # checking inputs, if anything is invalid an exit(1) will take place
        self.class_bn.fn_check_inputs(self.parameters)
        # checking inputs, if anything is invalid an exit(1) will take place
        self.fn_check_inputs_specific(self.parameters)

    def listing_parameter_values(self, in_logger, timer, title, in_config, given_parameter_values):
        timer.start()
        in_logger.info('=' * 50)
        in_logger.info(self.locale.gettext('{application_name} has started')
                       .replace('{application_name}', title))
        in_logger.info('~' * 50)
        in_logger.info(self.locale.gettext('Overview of input parameter given values'))
        in_logger.info('~' * 50)
        parameter_values_dictionary = given_parameter_values.__dict__
        for input_key, attributes in in_config.items():
            # checking first if short key was provided, otherwise consider longer
            if input_key in parameter_values_dictionary:
                key_value_to_consider = input_key
            else:
                key_value_to_consider = attributes['option_long'].replace('-', '_')
            # having the key consider we determine the value of the current parameter
            value_to_consider = parameter_values_dictionary[key_value_to_consider]
            # we build the parameter feedback considering "option_description"
            # and replacing %s with parameter value
            feedback = self.locale.gettext(attributes['option_description']) \
                .replace('%s', value_to_consider)
            # we finally write the feedback to logger
            in_logger.info(feedback)
        in_logger.info('~' * 50)
        timer.stop()



"""
BasicNeeds - useful functions library

This library has functions useful to keep main logic short and simple
"""
# package to handle date and times
from datetime import datetime, timedelta
# package to add support for multi-language (i18n)
import gettext
# package to handle files/folders and related metadata/operations
import os
# package regular expressions
import re


class BasicNeeds:
    locale = None

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    def fn_add_value_to_dictionary(self, in_list, adding_value, adding_type, reference_column):
        add_type = adding_type.lower()
        total_columns = len(in_list.copy())
        reference_indexes = {
            'add': {'after': 0, 'before': 0},
            'cycle_down_to': {'after': 0, 'before': 0}
        }
        if reference_column is not None:
            reference_indexes = {
                'add': {
                    'after': in_list.copy().index(reference_column) + 1,
                    'before': in_list.copy().index(reference_column),
                },
                'cycle_down_to': {
                    'after': in_list.copy().index(reference_column),
                    'before': in_list.copy().index(reference_column),
                }
            }
        positions = {
            'after': {
                'cycle_down_to': reference_indexes.get('cycle_down_to').get('after'),
                'add': reference_indexes.get('add').get('after'),
            },
            'before': {
                'cycle_down_to': reference_indexes.get('cycle_down_to').get('before'),
                'add': reference_indexes.get('add').get('before'),
            },
            'first': {
                'cycle_down_to': 0,
                'add': 0,
            },
            'last': {
                'cycle_down_to': total_columns,
                'add': total_columns,
            }
        }
        return self.add_value_to_dictionary_by_position({
            'adding_value': adding_value,
            'list': in_list.copy(),
            'position_to_add': positions.get(add_type).get('add'),
            'position_to_cycle_down_to': positions.get(add_type).get('cycle_down_to'),
            'total_columns': total_columns,
        })

    @staticmethod
    def add_value_to_dictionary_by_position(adding_dictionary):
        list_with_values = adding_dictionary['list']
        list_with_values.append(adding_dictionary['total_columns'])
        for counter in range(adding_dictionary['total_columns'],
                             adding_dictionary['position_to_cycle_down_to'], -1):
            list_with_values[counter] = list_with_values[(counter - 1)]
        list_with_values[adding_dictionary['position_to_add']] = adding_dictionary['adding_value']
        return list_with_values

    def fn_check_inputs(self, input_parameters):
        if input_parameters.output_log_file is not None:
            # checking log folder first as there's all further messages will be stored
            self.fn_validate_single_value(
                    os.path.dirname(input_parameters.output_log_file), 'folder')

    @staticmethod
    def fn_decide_by_omission_or_specific_true(in_dictionary, key_decision_factor):
        """
        Evaluates if a property is specified in a Dict structure

        @param in_dictionary: input Dict structure
        @param key_decision_factor: key used to search value in Dict structure
        @return: True|False
        """
        final_decision = False
        if key_decision_factor in in_dictionary:
            final_decision = True
        elif in_dictionary[key_decision_factor]:
            final_decision = True
        return final_decision

    @staticmethod
    def fn_evaluate_dict_values(in_dict):
        true_counted = 0
        for crt_value in in_dict:
            if in_dict[crt_value]:
                true_counted += 1
        all_true = False
        if true_counted == len(in_dict):
            all_true = True
        return all_true

    @staticmethod
    def fn_evaluate_list_values(in_list):
        true_counted = 0
        for crt_value in in_list:
            if crt_value:
                true_counted += 1
        all_true = False
        if true_counted == len(in_list):
            all_true = True
        return all_true

    def fn_final_message(self, local_logger, log_file_name, performance_in_seconds):
        total_time_string = str(timedelta(seconds=performance_in_seconds))
        if log_file_name == 'None':
            self.fn_timestamped_print(self.locale.gettext(
                'Application finished, whole script took {total_time_string}')
                                      .replace('{total_time_string}', total_time_string))
        else:
            local_logger.info(self.locale.gettext('Total execution time was {total_time_string}')
                              .replace('{total_time_string}', total_time_string))
            self.fn_timestamped_print(self.locale.gettext(
                'Application finished, for complete logged details please check {log_file_name}')
                                      .replace('{log_file_name}', log_file_name))

    @staticmethod
    def fn_multi_line_string_to_single(input_string):
        string_to_return = input_string.replace('\n', ' ').replace('\r', ' ')
        return re.sub(r'\s{2,100}', ' ', string_to_return).replace(' , ', ', ').strip()

    @staticmethod
    def fn_numbers_with_leading_zero(input_number_as_string, digits):
        final_number = input_number_as_string
        if len(input_number_as_string) < digits:
            final_number = '0' * (digits - len(input_number_as_string)) + input_number_as_string
        return final_number

    def fn_optional_print(self, boolean_variable, string_to_print):
        if boolean_variable:
            self.fn_timestamped_print(string_to_print)

    @staticmethod
    def fn_timestamped_print(string_to_print):
        print(datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S.%f %Z") + '- ' + string_to_print)

    def fn_validate_one_value(self, value_to_validate, validation_type):
        is_invalid = False
        message = ''
        if validation_type == 'file':
            is_invalid = (not os.path.isfile(value_to_validate))
            message = self.locale.gettext('Given file "{value_to_validate}" does not exist')
        elif validation_type == 'folder':
            is_invalid = (not os.path.isdir(value_to_validate))
            message = self.locale.gettext('Given folder "{value_to_validate}" does not exist')
        elif validation_type == 'url':
            url_reg_expression = 'https?://(?:www)?(?:[\\w-]{2,255}(?:\\.\\w{2,66}){1,2})'
            is_invalid = (not re.match(url_reg_expression, value_to_validate))
            message = self.locale.gettext('Given url "{value_to_validate}" is not valid')
        return {
            'is_invalid': is_invalid,
            'message': message.replace('{value_to_validate}', value_to_validate),
        }

    def fn_validate_single_value(self, value_to_validate, validation_type):
        validation_details = self.fn_validate_one_value(value_to_validate, validation_type)
        if validation_details['is_invalid']:
            self.fn_timestamped_print(validation_details['message'])
            exit(1)



"""
localization_setup - facilitates localization activities
"""
# facilitate dependencies management
from setuptools import setup
# facilitate internationalization
from babel.messages import frontend

setup(
    cmdclass={
        'compile_catalog': frontend.compile_catalog,
        'init_catalog': frontend.init_catalog,
        'update_catalog': frontend.update_catalog
    }
)



"""
TypeDetermination - a data type determination library

This library allows data type determination based on data frame content
"""
# package to add support for multi-language (i18n)
import gettext
# package to handle numerical structures
import numpy
# package to handle files/folders and related metadata/operations
import os
# regular expression package
import re
# package to facilitate common operations
from .BasicNeeds import BasicNeeds


class TypeDetermination(BasicNeeds):
    locale = None

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    def fn_analyze_field_content_to_establish_data_type(self, logger, field_characteristics,
                                                        data_types):
        crt_field_type = self.fn_type_determination(
                field_characteristics['unique_values'][0], data_types)
        # since date fields are not accepted to have ny null value by Tableau Hyper API
        # following forced String type is enforced
        if crt_field_type[:4] == 'date' and field_characteristics['nulls'] != 0:
            crt_field_type = 'str'
        # write aside the determined value
        field_dict = {
            'order': field_characteristics['order'],
            'name': field_characteristics['name'],
            'nulls': field_characteristics['nulls'],
            'panda_type': field_characteristics['panda_type'],
            'type': crt_field_type,
            'type_index': list(data_types.keys()).index(crt_field_type)
        }
        logger.debug(self.locale.gettext(
            'Column {column_order} having the name {column_name} '
            + 'has the value <{unique_values}> which means is of type "{field_type}"')
                     .replace('{column_order}', str(field_characteristics['order']))
                     .replace('{column_name}', field_characteristics['name'])
                     .replace('{unique_values}', str(field_characteristics['unique_values'][0]))
                     .replace('{field_type}', crt_field_type))
        if crt_field_type == 'str':
            return field_dict
        else:
            return self.fn_analyze_sample(logger, field_characteristics, field_dict, data_types)

    def fn_analyze_sample(self, logger, field_characteristics, field_structure, data_types):
        # Analyze unique values
        for unique_row_index, current_value in enumerate(field_characteristics['unique_values']):
            # determine the field type by current content
            crt_field_type = self.fn_type_determination(current_value, data_types)
            crt_type_index = list(data_types.keys()).index(crt_field_type)
            # is the current type is more important?
            if crt_type_index > field_structure['type_index']:
                logger.debug(self.locale.gettext(
                    'Column {column_order} having the name [{column_name}] '
                    + 'has the value <{current_value}> which means is of type "{column_type}" '
                    + 'and this is stronger than previously thought to be as "{column_type_old}"')
                             .replace('{column_order}', str(field_characteristics['order']))
                             .replace('{column_name}', field_characteristics['name'])
                             .replace('{current_value}', str(current_value))
                             .replace('{column_type}', crt_field_type)
                             .replace('{column_type_old}', field_structure['type']))
                field_structure['type'] = crt_field_type
                field_structure['type_index'] = crt_type_index
            # If currently determined field type is string makes not sense to scan any further
            if crt_field_type == 'str':
                return field_structure
        return field_structure

    def fn_detect_csv_structure(self, logger, input_csv_data_frame, input_parameters, data_types):
        col_idx = 0
        csv_structure = []
        # Cycle through all found columns
        for label, content in input_csv_data_frame.items():
            panda_determined_type = content.infer_objects().dtypes
            counted_nulls = content.isnull().sum()
            logger.debug(self.locale.gettext(
                'Field "{column_name}" according to Pandas package '
                + 'is of type "{panda_determined_type}" '
                + 'with {counted_nulls} counted NULLs')
                         .replace('{column_name}', label)
                         .replace('{panda_determined_type}', str(panda_determined_type))
                         .replace('{counted_nulls}', str(counted_nulls)))
            if panda_determined_type in ('float64', 'object'):
                list_unique_values = self.fn_unique_values_isolation(
                        logger, label, content, panda_determined_type, input_parameters)
                unique_v_list = {
                    'order': col_idx,
                    'name': label,
                    'nulls': counted_nulls,
                    'panda_type': panda_determined_type,
                    'unique_values': list_unique_values
                }
                str_unique_values = self.fn_multi_line_string_to_single(str(unique_v_list))
                logger.debug(self.locale.gettext(
                    'Unique list of values is: {unique_values_list}')
                             .replace('{unique_values_list}', str_unique_values))
                csv_structure.append(col_idx)
                csv_structure[col_idx] = \
                    self.fn_analyze_field_content_to_establish_data_type(
                            logger, unique_v_list, data_types)
            elif panda_determined_type in ('bool', 'int64'):
                csv_structure.append(col_idx)
                csv_structure[col_idx] = {
                    'order': col_idx,
                    'name': label,
                    'nulls': counted_nulls,
                    'panda_type': panda_determined_type,
                    'type': str(panda_determined_type).replace('64', ''),
                }
            col_idx += 1
        return csv_structure

    @staticmethod
    def fn_type_determination(input_variable_to_assess, data_types):
        # Website https://regex101.com/ was used to validate below code
        variable_to_assess = str(input_variable_to_assess)
        if variable_to_assess == '':
            return 'empty'
        else:
            for current_data_type, current_format in data_types.items():
                if re.match(current_format, variable_to_assess):
                    return current_data_type
            return 'str'

    def fn_unique_values_isolation(self, logger, label, content, panda_determined_type, in_prmtrs):
        counted_values_not_null = content.notnull().sum()
        counted_values_unique = content.nunique()
        content = content.dropna()
        if panda_determined_type == 'float64':
            content = content.apply(lambda x: x if (int(x) != x) else int(x))
        list_unique_values = content.unique()[0:int(in_prmtrs.unique_values_to_analyze_limit)]
        compact_unique_values = \
            self.fn_multi_line_string_to_single(
                '>, <'.join(numpy.array(list_unique_values, dtype=str)))
        logger.debug(self.locale.gettext(
            'Additional characteristics for the field "{column_name}" are: '
            + 'count of not-null values = {counted_values_not_null}, '
            + 'count of unique values = {counted_values_unique}, '
            + 'list of not-null and unique values is = <{compact_unique_values}>')
                     .replace('{column_name}', label)
                     .replace('{counted_values_not_null}', str(counted_values_not_null))
                     .replace('{counted_values_unique}', str(counted_values_unique))
                     .replace('{compact_unique_values}', str(compact_unique_values)))
        return list_unique_values



"""
localizations_common - common class for various localizations tasks
"""
import glob
# package to facilitate operating system project_locale detection
import locale
# package to handle files/folders and related metadata/operations
import os
import pathlib
# package to facilitate multiple operation system operations
import platform


class LocalizationsCommon:
    locale_implemented = [
        'it_IT',
        'ro_RO',
    ]
    operation_is_required = False

    def check_file_pairs(self, in_dict):
        get_details_to_operate = False
        file_situation_verdict = ''
        if os.path.lexists(pathlib.Path(in_dict['destination'])):
            source_last_modified = os.path.getmtime(in_dict['source'])
            compiled_last_modified = os.path.getmtime(in_dict['destination'])
            if source_last_modified > compiled_last_modified:
                self.operation_is_required = True
                get_details_to_operate = True
                print('#' + str(in_dict['counter']) + ', '
                      + 'For locale ' + in_dict['locale']
                      + ' the ' + in_dict['source file type name'] + ' file '
                      + os.path.basename(in_dict['source'])
                      + ' is newer than compiled one '
                      + os.path.basename(in_dict['destination'])
                      + ' therefore to remedy this, '
                      + in_dict['destination operation name'] + ' is required')
                print('===>' + in_dict['source'] + ' has ' + str(source_last_modified)
                      + ' vs. ' + str(compiled_last_modified))
                file_situation_verdict = 'newer'
        else:
            self.operation_is_required = True
            get_details_to_operate = True
            print('#' + str(in_dict['counter']) + ', '
                  + 'The file ' + os.path.basename(in_dict['source'])
                  + ' does not have compiled file for ' + in_dict['locale']
                  + ' therefore will require '
                  + in_dict['destination operation name'] + ' into following folder: '
                  + os.path.normpath(os.path.dirname(in_dict['destination'])))
            file_situation_verdict = 'missing'
        return get_details_to_operate, file_situation_verdict

    @staticmethod
    def file_counter_limit(in_file_counter, in_file_list_size):
        file_list_paring_complete = False
        if in_file_counter == in_file_list_size:
            file_list_paring_complete = True
        return file_list_paring_complete

    def get_region_language_to_use_from_operating_system(self):
        try:
            region_language_to_use = locale.getdefaultlocale('LC_ALL')
            if region_language_to_use[0] not in self.locale_implemented:
                language_to_use = self.locale_implemented[0]
            else:
                language_to_use = region_language_to_use[0]
        except AttributeError as err:
            print(err)
            language_to_use = self.locale_implemented[0]
        except ValueError as err:
            print(err)
            language_to_use = self.locale_implemented[0]
        return language_to_use

    @staticmethod
    def get_this_file_folder():
        return os.path.dirname(__file__)

    def get_project_localisation_source_files(self, in_extension):
        file_pattern = os.path.join(os.path.dirname(__file__), '**/*.' + in_extension)
        initial_list = glob.glob(file_pattern, recursive=True)
        normalizer = lambda x: self.path_normalize(x)
        return list(map(normalizer, initial_list))

    @staticmethod
    def get_project_root():
        file_parts = os.path.normpath(os.path.abspath(__file__))\
            .replace('\\', os.altsep).split(os.altsep)
        return os.altsep.join(file_parts[:-3])

    def get_virtual_environment_python_binary(self):
        python_binary = 'python'
        if platform.system() == 'Windows':
            project_root = self.get_project_root()
            join_separator = os.path.altsep
            elements_to_join = [
                project_root,
                'virtual_environment',
                'Scripts',
            ]
            virtual_env_long = join_separator.join(elements_to_join)
            elements_to_join[1] = 'venv'
            virtual_env_short = join_separator.join(elements_to_join)
            if os.path.isdir(virtual_env_long):
                python_binary = os.path.join(virtual_env_long, 'python.exe')
            elif os.path.isdir(virtual_env_short):
                python_binary = os.path.join(virtual_env_short, 'python.exe')
        return os.path.normpath(python_binary)

    def operate_localisation_files(self, in_dict_details_to_operate_with):
        if self.operation_is_required:
            for current_details_to_operate in in_dict_details_to_operate_with:
                os.system(self.get_virtual_environment_python_binary() + ' '
                          + os.path.join(os.path.dirname(__file__), 'localizations_setup.py')
                          + ' ' + current_details_to_operate['operation']
                          + ' --input-file=' + current_details_to_operate['input-file']
                          + ' --output-file=' + current_details_to_operate['output-file']
                          + ' --locale ' + current_details_to_operate['locale']
                          + current_details_to_operate['operation final flags'])

    @staticmethod
    def path_normalize(in_file_name):
        return os.path.join(os.path.normpath(os.path.dirname(in_file_name)),
                            os.path.basename(in_file_name))

    def run_localization_compile(self):
        command_parts_to_run = [
            self.get_virtual_environment_python_binary(),
            ' ',
            self.get_this_file_folder(),
            os.path.altsep,
            'localizations_compile.py',
        ]
        os.system(''.join(command_parts_to_run))




from datetime import datetime
import os
from sources.tableau_hyper_management.FileOperations import FileOperations
import unittest
# package to facilitate multiple operation system operations
import platform


class TestFileOperations(unittest.TestCase):

    def setUp(self) -> None:
        python_binary = 'python'
        if platform.system() == 'Windows':
            python_binary += '.exe'
        os.system(python_binary + ' '
                  + os.path.join(os.path.normpath(os.path.dirname(__file__))
                                 .replace('test', 'sources'), 'localizations_compile.py'))

    def test_file_statistics(self):
        class_fo = FileOperations()
        value_to_assert = class_fo.fn_get_file_statistics(__file__)['size [bytes]']
        value_to_compare_with = os.path.getsize(__file__)
        self.assertEqual(value_to_assert, value_to_compare_with)

    def test_file_dates(self):
        class_fo = FileOperations()
        value_to_assert = class_fo.fn_get_file_dates(__file__)['created']
        value_to_compare_with = datetime.fromtimestamp(os.path.getctime(__file__))
        self.assertEqual(value_to_assert, value_to_compare_with)



"""
localization_update - facilitates localization file maintenance
"""
import os

from localizations_common import LocalizationsCommon


class LocalizationsMaintainSources(LocalizationsCommon):

    def evaluate_update_necessity(self, in_list_localisation_source_files):
        list_size = len(in_list_localisation_source_files)
        self.operation_is_required = False
        file_list_paring_complete = False
        file_counter = 0
        domains_locale_to_update = []
        compiling_files_counter = 0
        while not file_list_paring_complete:
            template_localisation_file = in_list_localisation_source_files[file_counter]
            for current_locale in self.locale_implemented:
                join_separator = os.path.altsep
                elements_to_join = [
                    os.path.dirname(template_localisation_file),
                    current_locale,
                    'LC_MESSAGES',
                    os.path.basename(template_localisation_file).replace('.pot', '.po'),
                ]
                fn_dict = {
                    'destination': join_separator.join(elements_to_join),
                    'counter': file_counter,
                    'locale': current_locale,
                    'source': template_localisation_file,
                    'source file type name': 'template',
                    'destination operation name': 'update',
                }
                operation_check_result, file_situation_verdict = self.check_file_pairs(fn_dict)
                operation_to_execute = ''
                operation_final_flags = ''
                if 'missing' in file_situation_verdict:
                    operation_to_execute = 'init_catalog'
                elif 'newer' in file_situation_verdict:
                    operation_to_execute = 'update_catalog'
                    operation_final_flags = ' --previous'
                if operation_check_result:
                    domains_locale_to_update.append(compiling_files_counter)
                    domains_locale_to_update[compiling_files_counter] = {
                        'input-file': fn_dict['source'],
                        'operation': operation_to_execute,
                        'operation final flags': operation_final_flags,
                        'output-file': fn_dict['destination'],
                        'locale': fn_dict['locale'],
                    }
                    compiling_files_counter += 1
            file_counter += 1
            file_list_paring_complete = self.file_counter_limit(file_counter, list_size)
        return domains_locale_to_update


my_class = LocalizationsMaintainSources()

locale_source_files = my_class.get_project_localisation_source_files('pot')
operation_locale_dict = my_class.evaluate_update_necessity(locale_source_files)
print(operation_locale_dict)
my_class.operate_localisation_files(operation_locale_dict)



"""
TableauHyperApiExtraLogic - a Hyper client library.

This library allows packaging CSV content into HYPER format with data type checks
"""
# package to add support for multi-language (i18n)
import gettext
# package to handle files/folders and related metadata/operations
import os
# package regular expression
import re
# package to handle numerical structures
import numpy
# package to handle Data Frames (in this file)
import pandas as pd
# Custom classes from Tableau Hyper package
from tableauhyperapi import HyperProcess, Telemetry, \
    Connection, CreateMode, \
    NOT_NULLABLE, NULLABLE, SqlType, TableDefinition, \
    Inserter, \
    TableName, \
    HyperException


class TableauHyperApiExtraLogic:
    locale = None

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    def fn_build_hyper_columns_for_csv(self, logger, timer, detected_csv_structure):
        timer.start()
        list_to_return = []
        for current_field_structure in detected_csv_structure:
            list_to_return.append(current_field_structure['order'])
            current_column_type = self.fn_convert_to_hyper_types(current_field_structure['type'])
            logger.debug(self.locale.gettext(
                'Column {column_order} having name [{column_name}] and type "{column_type}" '
                + 'will become "{column_type_new}"')
                         .replace('{column_order}', str(current_field_structure['order']))
                         .replace('{column_name}', current_field_structure['name'])
                         .replace('{column_type}', current_field_structure['type'])
                         .replace('{column_type_new}', str(current_column_type)))
            nullability_value = NULLABLE
            if current_field_structure['nulls'] == 0:
                nullability_value = NOT_NULLABLE
            list_to_return[current_field_structure['order']] = TableDefinition.Column(
                name=current_field_structure['name'],
                type=current_column_type,
                nullability=nullability_value
            )
        logger.info(self.locale.gettext('Building Hyper columns completed'))
        timer.stop()
        return list_to_return

    @staticmethod
    def fn_convert_to_hyper_types(given_type):
        switcher = {
            'empty': SqlType.text(),
            'bool': SqlType.bool(),
            'int': SqlType.big_int(),
            'float-dot': SqlType.double(),
            'date-YMD': SqlType.date(),
            'date-MDY': SqlType.date(),
            'date-DMY': SqlType.date(),
            'time-24': SqlType.time(),
            'time-12': SqlType.time(),
            'datetime-24-YMD': SqlType.timestamp(),
            'datetime-12-MDY': SqlType.timestamp(),
            'datetime-24-DMY': SqlType.timestamp(),
            'str': SqlType.text()
        }
        identified_type = switcher.get(given_type)
        if identified_type is None:
            identified_type = SqlType.text()
        return identified_type

    def fn_create_hyper_file_from_csv(self, local_logger, timer, input_csv_data_frame,
                                      in_data_type, given_parameters):
        hyper_cols = self.fn_build_hyper_columns_for_csv(local_logger, timer, in_data_type)
        # The rows to insert into the <hyper_table> table.
        data_to_insert = self.fn_rebuild_csv_content_for_hyper(
                local_logger, timer, input_csv_data_frame, in_data_type)
        # Starts the Hyper Process with telemetry enabled/disabled to send data to Tableau or not
        # To opt in, simply set telemetry=Telemetry.SEND_USAGE_DATA_TO_TABLEAU.
        # To opt out, simply set telemetry=Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU.
        with HyperProcess(telemetry=Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU) as hyper:
            # Creates new Hyper file <output_hyper_file>
            # Replaces file with CreateMode.CREATE_AND_REPLACE if it already exists.
            timer.start()
            with Connection(endpoint=hyper.endpoint,
                            database=given_parameters.output_file,
                            create_mode=CreateMode.CREATE_AND_REPLACE) as hyper_connection:
                local_logger.info('Connection to the Hyper engine '
                                  + f'file "{given_parameters.output_file}" has been created.')
                timer.stop()
                schema_name = 'Extract'
                self.fn_create_hyper_schema(local_logger, timer, hyper_connection, schema_name)
                hyper_table_name = 'Extract'
                hyper_table = self.fn_create_hyper_table(local_logger, timer, {
                    'columns': hyper_cols,
                    'connection': hyper_connection,
                    'schema name': schema_name,
                    'table name': hyper_table_name,
                })
                self.fn_insert_data_into_hyper_table(local_logger, timer, {
                    'connection': hyper_connection,
                    'data': data_to_insert,
                    'table': hyper_table,
                })
                self.fn_get_records_count_from_table(local_logger, timer, {
                    'connection': hyper_connection,
                    'table name': hyper_table_name,
                })
            local_logger.info(self.locale.gettext(
                'Connection to the Hyper engine file has been closed'))
        local_logger.info(self.locale.gettext('Hyper engine process has been shut down'))

    def fn_insert_data_into_hyper_table(self, local_logger, timer, in_dict):
        timer.start()
        # Execute the actual insert
        with Inserter(in_dict['connection'], in_dict['table']) as hyper_insert:
            hyper_insert.add_rows(rows=in_dict['data'])
            hyper_insert.execute()
        local_logger.info(self.locale.gettext('Data has been inserted into Hyper table'))
        timer.stop()

    def fn_create_hyper_schema(self, local_logger, timer, in_hyper_connection, in_schema_name):
        timer.start()
        in_hyper_connection.catalog.create_schema(in_schema_name)
        local_logger.info(self.locale.gettext(
            'Hyper schema "{hyper_schema_name}" has been created')
                          .replace('{hyper_schema_name}', in_schema_name))
        timer.stop()

    def fn_create_hyper_table(self, local_logger, timer, in_dict):
        timer.start()
        out_hyper_table = TableDefinition(
            TableName(in_dict['schema name'], in_dict['table name']),
            columns=in_dict['columns'],
        )
        in_dict['connection'].catalog.create_table(table_definition=out_hyper_table)
        local_logger.info(self.locale.gettext(
            'Hyper table "{hyper_table_name}" has been created')
                          .replace('{hyper_table_name}', in_dict['table name']))
        timer.stop()
        return out_hyper_table

    def fn_get_records_count_from_table(self, local_logger, timer, in_dict):
        timer.start()
        # Number of rows in the <hyper_table> table.
        # `execute_scalar_query` is for executing a query
        # that returns exactly one row with one column.
        query_to_run = 'SELECT COUNT(*) FROM {hyper_table_name}'\
            .replace('{hyper_table_name}', in_dict['table name'])
        row_count = in_dict['connection'].execute_scalar_query(query=query_to_run)
        local_logger.info(self.locale.gettext(
            'Table {hyper_table_name} has {row_count} rows')
                          .replace('{hyper_table_name}', in_dict['table name'])
                          .replace('{row_count}', row_count))
        timer.stop()

    def fn_rebuild_csv_content_for_hyper(self, in_logger, timer, input_df, detected_fields_type):
        timer.start()
        input_df.replace(to_replace=[numpy.nan], value=[None], inplace=True)
        # Cycle through all found columns
        for current_field in detected_fields_type:
            fld_nm = current_field['name']
            in_logger.debug(self.locale.gettext(
                'Column {column_name} has pandas data type = {column_pandas_type} '
                + 'and python data type = {column_python_type}')
                            .replace('{column_name}', current_field['name'])
                            .replace('{column_pandas_type}', str(current_field['panda_type']))
                            .replace('{column_python_type}', str(current_field['type'])))
            input_df[fld_nm] = self.fn_reevaluate_single_column(input_df, fld_nm, current_field)
            if current_field['panda_type'] == 'object' \
                and current_field['type'] == 'int':
                input_df[fld_nm] = input_df[fld_nm].fillna(0).astype('int64')
                in_logger.debug(self.locale.gettext(
                    'Column {column_name} has been forced converted to {forced_type}')
                                .replace('{column_name}', current_field['name'])
                                .replace('{forced_type}', 'Int'))
            if str(current_field['type']) == 'float-dot':
                input_df[fld_nm] = input_df[fld_nm].astype(float)
                in_logger.debug(self.locale.gettext(
                    'Column {column_name} has been forced converted to {forced_type}')
                                .replace('{column_name}', current_field['name'])
                                .replace('{forced_type}', 'Float'))
        in_logger.info(self.locale.gettext(
            'Re-building CSV content for maximum Hyper compatibility has been completed'))
        timer.stop()
        return input_df.values

    def fn_reevaluate_single_column(self, given_df, given_field_name, current_field_details):
        if current_field_details['type'] == 'str':
            given_df[given_field_name] = given_df[given_field_name].astype(str)
        elif current_field_details['panda_type'] == 'float64' \
                and current_field_details['type'] == 'int':
            given_df[given_field_name] = given_df[given_field_name].fillna(0).astype('int64')
        elif current_field_details['type'][0:5] in ('date-', 'datet', 'time-'):
            given_df[given_field_name] = self.fn_string_to_date(given_field_name, given_df)
        return given_df[given_field_name]

    def fn_run_hyper_creation(self, local_logger, timer, input_data_frame, input_data_type,
                              given_parameters):
        try:
            self.fn_create_hyper_file_from_csv(local_logger, timer, input_data_frame,
                                               input_data_type, given_parameters)
        except HyperException as ex:
            local_logger.error(str(ex).replace(chr(10), ' '))
            exit(1)

    @staticmethod
    def fn_string_to_date(in_col_name, in_data_frame):
        if re.match('-YMD', in_col_name):
            in_data_frame[in_col_name] = pd.to_datetime(in_data_frame[in_col_name], yearfirst=True)
        elif re.match('-DMY', in_col_name):
            in_data_frame[in_col_name] = pd.to_datetime(in_data_frame[in_col_name], dayfirst=True)
        else:
            in_data_frame[in_col_name] = pd.to_datetime(in_data_frame[in_col_name])
        return in_data_frame[in_col_name]



"""
setup - ensures proper package setup

This file is ensuring proper package setup is performed to ensure all prerequisites are satisfied 
and correct execution is possible
"""
# package to handle files/folders and related metadata/operations
import os
# package to facilitate multiple operation system operations
import platform
# facilitate dependencies management
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.md'), 'r') as fh:
    long_description_readme = fh.read()

this_package_website = 'https://github.com/danielgp/tableau-hyper-management'

setup(
    author='Daniel Popiniuc',
    author_email='danielpopiniuc@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],
    description='Wrapper to ease data management into Tableau Hyper format from CSV files',
    include_package_data=True,
    install_requires=[
        'Babel>=2.8.0,<3',
        'cffi>=1.13.2,<2',
        'codetiming>=1.1,<2',
        'numpy>=1.17.4,<2',
        'pandas>=0.25.3,<1',
        'tableauhyperapi',
        'tableauserverclient',
        'xlrd>=1,<2',
        'xlsxwriter>=1,<2'
    ],
    keywords=[
        'tableau',
        'hyper',
        'csv',
        'json',
        'excel',
        'pickle'
    ],
    license='LGPL3',
    long_description=long_description_readme,
    long_description_content_type='text/markdown',
    name='tableau-hyper-management',
    packages=find_packages('tableau_hyper_management'),
    project_urls={
        'Documentation': this_package_website + '/blob/master/README.md',
        'Issue Tracker': this_package_website +
                         '/issues?q=is%3Aissue+is%3Aopen+sort%3Aupdated-desc',
        'Source Code': this_package_website
    },
    python_requires='>=3.6',
    url=this_package_website + '/releases',  # project home page, if any
    version='1.3.13'
)



"""
main - entry point of the package

This file is performing CSV read into HYPER file and measures time elapsed (performance)
"""
# package to handle files/folders and related metadata/operations
import os
# Custom classes specific to this package
from project_locale.localizations_common import LocalizationsCommon
from tableau_hyper_management.ProjectNeeds import ProjectNeeds
from tableau_hyper_management.TableauHyperApiExtraLogic import TableauHyperApiExtraLogic
from tableau_hyper_management.TypeDetermination import TypeDetermination
# get current script name
SCRIPT_NAME = os.path.basename(__file__).replace('.py', '')

# main execution logic
if __name__ == '__main__':
    # instantiate Localizations Common class
    class_lc = LocalizationsCommon()
    # ensure all compiled localization files are in place (as needed for localized messages later)
    class_lc.run_localization_compile()
    # establish localization language to use
    language_to_use = class_lc.get_region_language_to_use_from_operating_system()
    # instantiate Extractor Specific Needs class
    class_pn = ProjectNeeds(SCRIPT_NAME, language_to_use)
    # load application configuration (inputs are defined into a json file)
    class_pn.load_configuration()
    # adding a special case data type
    class_pn.config['data_types']['empty'] = '^$'
    class_pn.config['data_types']['str'] = ''
    # initiate Logging sequence
    class_pn.initiate_logger_and_timer()
    # reflect title and input parameters given values in the log
    class_pn.class_clam.listing_parameter_values(
        class_pn.class_ln.logger, class_pn.timer, 'Tableau Hyper Converter',
        class_pn.config['input_options'][SCRIPT_NAME], class_pn.parameters)
    relevant_files_list = class_pn.class_fo.fn_build_file_list(
            class_pn.class_ln.logger, class_pn.timer, class_pn.parameters.input_file)
    # log file statistic details
    class_pn.class_fo.fn_store_file_statistics(
            class_pn.class_ln.logger, class_pn.timer, relevant_files_list, 'Input')
    # loading from a specific folder all files matching a given pattern into a data frame
    input_dict = {
        'compression': class_pn.parameters.input_file_compression,
        'field delimiter': class_pn.parameters.csv_field_separator,
        'file list': relevant_files_list,
        'format': class_pn.parameters.input_file_format,
        'name': 'irrelevant',
    }
    working_data_frame = class_pn.class_dio.fn_load_file_into_data_frame(
            class_pn.class_ln.logger, class_pn.timer, input_dict)
    if working_data_frame is not None:
        if class_pn.parameters.output_file_format.lower() in ('csv', 'excel', 'pickle'):
            output_dict = input_dict
            output_dict['file list'] = 'irrelevant'
            output_dict['format'] = class_pn.parameters.output_file_format
            output_dict['name'] = class_pn.parameters.output_file
            output_dict['compression'] = class_pn.parameters.output_file_compression
            class_pn.class_dio.fn_store_data_frame_to_file(
                class_pn.class_ln.logger, class_pn.timer, working_data_frame, output_dict)
            # store statistics about output file
            class_pn.class_fo.fn_store_file_statistics(
                class_pn.class_ln.logger, class_pn.timer,
                class_pn.parameters.output_file, 'Generated')
        elif class_pn.parameters.output_file_format.lower() == 'hyper':
            if class_pn.parameters.input_file_format.lower() == 'csv':
                class_pn.timer.start()
                c_td = TypeDetermination()
                # advanced detection of data type within Data Frame
                detected_csv_structure = c_td.fn_detect_csv_structure(
                    class_pn.class_ln.logger, working_data_frame, class_pn.parameters,
                    class_pn.config['data_types'])
                class_pn.timer.stop()
                # instantiate Tableau Hyper Api Extra Logic class
                class_thael = TableauHyperApiExtraLogic()
                # create HYPER from Data Frame
                class_thael.fn_run_hyper_creation(
                    class_pn.class_ln.logger, class_pn.timer, working_data_frame,
                    detected_csv_structure, class_pn.parameters)
                # store statistics about output file
                class_pn.class_fo.fn_store_file_statistics(
                    class_pn.class_ln.logger, class_pn.timer,
                    class_pn.parameters.output_file, 'Generated')
            else:
                print('For time being only CSV file types are supported as input file type '
                      + 'in combination with Hyper as output file type.'
                      + 'An enhanced version will be deployed in near future!')
    # just final message
    class_pn.class_bn.fn_final_message(
            class_pn.class_ln.logger, class_pn.parameters.output_log_file,
            class_pn.timer.timers.total(SCRIPT_NAME))
